
import os

from ..buildtool import BuildTool
from ..testtool import TestTool
from .piptoolmixin import PipToolMixIn

class setuptoolsTool( PipToolMixIn, BuildTool, TestTool ):
    """Easily download, build, install, upgrade, and uninstall Python packages.
    
Home: https://pypi.python.org/pypi/setuptools

Not assumed to be used directly.
"""    
    def autoDetect( self, config ) :
        return self._autoDetectByCfg(
                config,
                [ 'setup.py' ]
        )
    
    def uninstallTool( self, env ):
        pass
        
    def initEnv( self, env ):
        virtualenv_dir = env['virtualenvDir']
        self._have_tool = os.path.exists(os.path.join(virtualenv_dir, 'bin', 'easy_install'))
        
    def onPrepare( self, config ):
        for d in ['build', 'dist']:
            if os.path.exists(d):
                self._rmTree(d)

    def onBuild( self, config ):
        env = config['env']
        self._requirePip( env, 'wheel' )
        self._callExternal( [ env['pythonBin'], 'setup.py', 'sdist', 'bdist_wheel' ] )
    

    def onCheck( self, config ):
        env = config['env']
        self._requirePip( env, 'docutils' )
        self._requirePip( env, 'readme' )
        self._callExternal( [ env['pythonBin'], 'setup.py', 'check', '-mrs' ] )

    



import os

from ..runenvtool import RunEnvTool
from .bashtoolmixin import BashToolMixIn

class virtualenvTool( BashToolMixIn, RunEnvTool ):
    """Virtual Python Environment builder.
    
Home: https://pypi.python.org/pypi/virtualenv
"""
    def getDeps( self ) :
        return [ 'bash', 'python' ]
    
    def getPostDeps( self ) :
        # we need to ensure that if cid is called from virtualenv env
        # it's still can be used
        # So, CID can be in system and in each virtualenv as well
        return ['cid']
    
    def envNames( self ) :
        return [ 'virtualenvDir', 'virtualenvVer']
    
    def _installTool( self, env ):
        python_ver = env['pythonVer'].split('.')
        
        virtualenv_dir = env['virtualenvDir']
        
        # virtualenv depends on ensurepip and provides no setuptools
        # ensurepip is not packaged on all OSes...
        if False and int(python_ver[0]) == 3 and int(python_ver[1]) >= 3:
            self._callExternal([
                env['pythonRawBin'],
                '-m', 'venv',
                '--system-site-packages',
                '--symlinks',
                '--without-pip', # avoid ensurepip run
                '--clear',
                virtualenv_dir
            ])
        else:
            # Looks quite stupid, but it's a workaround for different OSes
            pip = self._which('pip')
            
            if not pip:
                self._requirePackages(['python-virtualenv'])
                self._requireEmerge(['dev-python/virtualenv'])
                self._requirePacman(['python-virtualenv'])
                
                self._requireDeb(['python-pip'])
                self._requireYum(['python2-pip'])
                self._requireEmerge(['dev-python/pip'])
                self._requirePacman(['python-pip'])
                
                pip = self._which('pip')
                
            if pip:
                self._trySudoCall([ pip, 'install', '-q', '--upgrade', 'virtualenv>={0}'.format(env['virtualenvVer']) ])
            
            virtualenv = self._which('virtualenv')
            
            if not virtualenv:
                raise RuntimeError('Failed to find virtualenv')

            self._callExternal([
                virtualenv,
                '--python={0}'.format(env['pythonRawBin']),
                '--clear',
                virtualenv_dir
            ])
            
    def updateTool( self, env ):
        pass
    
    def initEnv( self, env ) :
        virtualenv_dir = '.virtualenv-{0}'.format(env['pythonVer'][:3])
        virtualenv_dir = env.setdefault('virtualenvDir', os.path.join(os.environ['HOME'], virtualenv_dir))
        
        env.setdefault('virtualenvVer', '15.1.0')
        
        self._have_tool = os.path.exists(os.path.join(virtualenv_dir, 'bin', 'activate'))
        
        if self._have_tool:
            env_to_set = self._callBash( env,
                'source {0}/bin/activate && env | grep \'{0}\''.format(virtualenv_dir),
                verbose = False
            )
                
            self._updateEnvFromOutput(env_to_set)
            # reverse-dep hack
            env['pythonBin'] = os.path.join(virtualenv_dir, 'bin', 'python')




#!/usr/bin/env python
"""FutoIn Continuous Integration & Delivery Tool.

Usage:
    cid tag <branch> [<next_version>] [--vcsRepo vcs_url] [--wcDir wc_dir]
    cid prepare [<vcs_ref>] [--vcsRepo vcs_url] [--wcDir wc_dir]
    cid build
    cid package
    cid check [--permissive]
    cid promote <package> <rms_pool> [--rmsRepo rms_url] [--rmsHash type_value]
    cid deploy vcstag [<vcs_ref>] [--vcsRepo vcs_url] [--redeploy] [--deployDir deploy_dir]    
    cid deploy vcsref <vcs_ref> [--vcsRepo vcs_url] [--redeploy] [--deployDir deploy_dir]    
    cid deploy [rms] <rms_pool> [<package>] [--rmsRepo rms_url] [--rmsHash type_value] [--redeploy] [--deployDir deploy_dir] [--build]
    cid run [<command>]
    cid ci_build <vcs_ref> <rms_pool> [--vcsRepo vcs_url] [--rmsRepo rms_url] [--permissive]
    cid tool exec <tool_name> [-- <tool_arg>...]
    cid tool (install|uninstall|update|test|env) [<tool_name>]
    cid tool (prepare|build|check|package|migrate) <tool_name>
    cid tool list
    cid tool describe <tool_name>

Options:
    --vcsRepo vcs_type:vcs_url
    --rmsRepo rms_type:rms_url
    --wcDir wc_dir
    --rmsHash hash_type:value
"""

from __future__ import print_function, absolute_import

from .cidtool import CIDTool

try:
    from docopt import docopt
except ImportError:
    # fallback to "hardcoded"
    from futoin.cid.contrib.docopt import docopt

import os, sys

if sys.version_info < (2,7):
    print( 'Sorry, but only Python version >= 2.7 is supported!' )
    sys.exit( 1 )

from . import __version__ as version
__all__ = ['run']

def run():
    args = docopt( __doc__, version='FutoIn CIDTool v{0}'.format(version) )

    if type(args) == str:
        print(args)
    else:
        overrides = {}
        #---
        vcsArg = args.get('--vcsRepo', None)

        if vcsArg:
            (overrides['vcs'],
             overrides['vcsRepo']) = vcsArg.split(':', 1)
        #---
        rmsArg = args.get('--rmsRepo', None)

        if rmsArg:
            (overrides['rms'],
             overrides['rmsRepo']) = rmsArg.split(':', 1)
        overrides['rmsHash'] = args['--rmsHash']
        overrides['rmsPool'] = args['<rms_pool>']
        #---
        if args['ci_build'] :
            def_wc_dir = 'ci_build'
        else :
            def_wc_dir = '.'
        overrides['wcDir'] = os.path.realpath(args['--wcDir'] or def_wc_dir)
        #--
        deploy_dir = args['--deployDir']
        overrides['deployDir'] = deploy_dir and os.path.realpath(deploy_dir) or None
        overrides['reDeploy'] = args['--redeploy'] and True or False
        if args['--build']:
            overrides['deployBuild'] = True
        
        #---
        overrides['vcsRef'] = args['<vcs_ref>']

        #---
        tool = args['<tool_name>']
        overrides['tool'] = tool
        overrides['toolTest'] = args['test'] or args['uninstall'] or args['describe']
        
        #---
        if args['--permissive']:
            overrides['permissiveChecks'] = True
        
        #---
        cit = CIDTool( overrides = overrides )
        
        if args['tag'] :
            cit.tag( args['<branch>'], args['<next_version>'] )
        elif args['tool'] :
            if args['exec']:
                cit.tool_exec( tool, args['<tool_arg>'] )
            elif args['list']:
                cit.tool_list()
            else:
                subcmds = [
                    'install',
                    'uninstall',
                    'update',
                    'test',
                    'env',
                    'prepare',
                    'build',
                    'check',
                    'package',
                    'migrate',
                    'describe',
                ]
                for cmd in subcmds:
                    if args[cmd]:
                        getattr( cit, 'tool_'+cmd )( tool )
                        break
                else:
                    print( "Unknown Command" )
                    sys.exit( 1 )
        elif args['prepare'] :
            cit.prepare( args['<vcs_ref>'] )
        elif args['build'] :
            cit.build()
        elif args['package']:
            cit.package()
        elif args['check'] :
            cit.check()
        elif args['promote'] :
            cit.promote( args['<package>'], args['<rms_pool>'] )
        elif args['deploy'] :
            if args['vcsref']:
                overrides['deployBuild'] = True
                cit.deploy( 'vcsref', args['<vcs_ref>'] )
            elif args['vcstag']:
                overrides['deployBuild'] = True
                cit.deploy( 'vcstag', args['<vcs_ref>'] )
            else :
                cit.deploy( 'rms', args['<rms_pool>'], args['<package>'] )
        elif args['run'] :
            cit.run( args['<command>'] or 'start' )
        elif args['ci_build'] :
            cit.ci_build( args['<vcs_ref>'], args['<rms_pool>'] )
        else:
            print( "Unknown Command" )
            sys.exit( 1 )




import os

from ..vcstool import VcsTool
from .bashtoolmixin import BashToolMixIn

class gitTool( BashToolMixIn, VcsTool ):
    """Git distributed version control system.
    
Home: https://git-scm.com/

Auto-detected based on .git folder.

Git tool forcibly sets user.email and user.name, 
if not set by user.
"""    
    _rev = None
    
    def getDeps(self):
        return ['bash', 'tar']

    def autoDetect( self, config ) :
        return self._autoDetectVCS( config, '.git' )
    
    def _installTool( self, env ):
        self._requirePackages(['git'])
        self._requireEmerge(['dev-vcs/git'])
        self._requirePacman(['git'])

    def _checkGitConfig( self, env ):
        gitBin = env['gitBin']
        user_email = None
        user_name = None
        
        try:
            user_email = self._callExternal([
                gitBin, 'config', 'user.email',
            ], verbose=False).strip()
        except:
            pass
        
        try:
            user_name = self._callExternal([
                gitBin, 'config', 'user.name',
            ], verbose=False).strip()
        except:
            pass
        
        if not user_email:
            self._callExternal([
                gitBin, 'config', 'user.email',
                env.get('gitUserEmail', 'noreply@futoin.org')
            ])

        if not user_name:
            self._callExternal([
                gitBin, 'config', 'user.name',
                env.get('gitUserName', 'FutoIn CITool')
            ])

    def vcsGetRepo( self, config, wc_dir=None ):
        git_dir = wc_dir or os.path.join(config['wcDir'], '.git')
        
        return self._callExternal( [
            config['env']['gitBin'],
            '--git-dir={0}'.format(git_dir),
            'config',
            '--get',
            'remote.origin.url'
        ] ).strip()
    
    def _gitCompareRepo( self, cfg, act ):
        return cfg == act or ('ssh://'+cfg) == act
    
    def vcsCheckout( self, config, vcs_ref ):
        gitBin = config['env']['gitBin']
        wc_dir = config['wcDir']
        vcsRepo = config['vcsRepo']
        
        if os.path.isdir( '.git' ):
            remote_url = self.vcsGetRepo( config, '.git')
            
            if not self._gitCompareRepo( vcsRepo, remote_url ) :
                raise RuntimeError( "Git remote mismatch: {0} != {1}"
                    .format(vcsRepo, remote_url ) )

            self._callExternal( [ gitBin, 'fetch', '-q'  ] )
        else :
            self._callExternal( [ gitBin, 'clone', '-q', vcsRepo, wc_dir ] )
            
        remote_branch = self._callExternal( [ gitBin, 'branch', '-q', '--list', 'origin/'+vcs_ref ] ).strip()
            
        if self._callExternal( [ gitBin, 'branch', '-q', '--list', vcs_ref ] ).strip():
            self._callExternal( [ gitBin, 'checkout', '-q', vcs_ref ] )

            if remote_branch:
                self._callExternal( [ gitBin, 'branch', '-q', '--set-upstream-to', 'origin/'+vcs_ref ] )
                self._callExternal( [ gitBin, 'rebase', 'origin/'+vcs_ref  ] )
        elif remote_branch:
            self._callExternal( [ gitBin, 'checkout', '-q', '--track', '-b', vcs_ref, 'origin/'+vcs_ref ] )
        else:
            self._callExternal( [ gitBin, 'checkout', '-q', vcs_ref ] )
    
    def vcsCommit( self, config, message, files ):
        env = config['env']
        gitBin = env['gitBin']
        self._checkGitConfig(env)
        self._callExternal( [ gitBin, 'commit', '-q', '-m', message ] + files )
    
    def vcsTag( self, config, tag, message ):
        gitBin = config['env']['gitBin']
        self._callExternal( [ gitBin, 'tag', '-a', '-m', message, tag ] )
    
    def vcsPush( self, config, refs ):
        gitBin = config['env']['gitBin']
        self._callExternal( [ gitBin, 'push', '-q', 'origin' ] + refs )
        
    def vcsGetRevision( self, config ) :
        gitBin = config['env']['gitBin']
        return self._callExternal( [ gitBin, 'rev-parse', 'HEAD' ] ).strip()
    
    def vcsGetRefRevision( self, config, vcs_cache_dir, branch ) :
        res = self._callExternal( [
            config['env']['gitBin'],
            'ls-remote', '--refs',
            config['vcsRepo'],
            'refs/heads/{0}'.format(branch)
        ] ).strip()
        
        if res:
            self._rev = res.split()[0]
            return self._rev
        
        raise RuntimeError( "Uknown Git ref: {0}".format( branch ) )
    

    def vcsListTags( self, config, vcs_cache_dir, tag_hint ) :
        if tag_hint:
            tag_hint = ['refs/tags/{0}'.format(tag_hint)]
        else:
            tag_hint = []
        
        res = self._callExternal( [
            config['env']['gitBin'],
            'ls-remote','--tags', '--refs',
            config['vcsRepo']
        ] + tag_hint ).strip().split("\n")

        return [ v.split()[1].replace('refs/tags/', '') for v in res ]

    def vcsExport( self, config, vcs_cache_dir, vcs_ref, dst_path ) :
        env = config['env']
        gitBin = env['gitBin']
        vcsRepo = config['vcsRepo']

        if os.path.exists(vcs_cache_dir):
            remote_url = self.vcsGetRepo( config, vcs_cache_dir )
            
            if self._gitCompareRepo(vcsRepo,  remote_url):
                print('WARNING: removing git cache on remote URL mismatch: {0} != {1}'
                      .format(remote_url, vcsRepo))
                self._rmTree(vcs_cache_dir)
        
        if not os.path.exists(vcs_cache_dir):
             self._callExternal( [
                env['gitBin'],
                'clone', '--mirror',
                vcsRepo,
                vcs_cache_dir
            ] )
        else:
            self._callExternal( [
                env['gitBin'],
                '--git-dir={0}'.format(vcs_cache_dir),
                'fetch'
            ] )
            
        if os.path.exists(dst_path):
            self._rmTree(dst_path)

        os.mkdir(dst_path)
        
        if self._rev:
            vcs_ref = self._rev
        
        self._callBash(env, '{0} --git-dir={1} archive --format=tar {2} | {3} x -C {4}'
                .format(config['env']['gitBin'], vcs_cache_dir, vcs_ref, env['tarBin'], dst_path))





import os, re
import xml.dom.minidom

from ..vcstool import VcsTool
from ..rmstool import RmsTool

class svnTool( VcsTool, RmsTool ):
    """Apache Subversion: Enterprise-class centralized version control for the masses.
    
Home: https://subversion.apache.org/
"""    
    _rev = None
    
    def _installTool( self, env ):
        self._requirePackages(['subversion'])
        self._requireEmerge(['dev-vcs/subversion'])
        self._requirePacman(['subversion'])
    
    def autoDetect( self, config ) :
        return ( self._autoDetectVCS( config, '.svn' )
                or RmsTool.autoDetect( self, config ) )
    
    def vcsGetRepo( self, config, wc_dir=None ):
        svn_info = self._callExternal([
            config['env']['svnBin'], 'info', '--xml', wc_dir or config['wcDir']
        ])
        svn_info = xml.dom.minidom.parseString( svn_info )
        svn_info = svn_info.getElementsByTagName('url')
        url = svn_info[0].firstChild.nodeValue
        url = re.sub( '/(trunk|branches|tags).+$', '', url )
        return url
    
    def vcsCheckout( self, config, vcs_ref ):
        env = config['env']
        svnBin = env['svnBin']
        wc_dir = config['wcDir']
            
        branch_path = '%s/branches/%s' % ( config['vcsRepo'], vcs_ref )
        tag_path = '%s/tags/%s' % ( config['vcsRepo'], vcs_ref )
        
        if self._callExternal( [ svnBin, 'ls', branch_path ], suppress_fail=True ) :
            svn_repo_path = branch_path
        elif self._callExternal( [ svnBin, 'ls', tag_path ], suppress_fail=True ) :
            svn_repo_path = tag_path
        else:
            raise RuntimeError( "VCS ref was not found: " + vcs_ref )

        if os.path.isdir( '.svn' ):
            self._callExternal( [ svnBin, 'switch', svn_repo_path ] )
        else:
            self._callExternal(
                [ svnBin, 'checkout',
                 svn_repo_path,
                 wc_dir ] )
    
    def vcsCommit( self, config, message, files ):
        svnBin = config['env']['svnBin']
        self._callExternal(
                [ svnBin, 'commit',
                 '-m', message ] + files )
    
    def vcsTag( self, config, tag, message ):
        env = config['env']
        svnBin = env['svnBin']

        svn_info = self._callExternal([
            config['env']['svnBin'], 'info', '--xml'
        ])
        svn_info = xml.dom.minidom.parseString( svn_info )
        svn_info = svn_info.getElementsByTagName('url')
        svn_url = svn_info[0].firstChild.nodeValue

        self._callExternal( [
            svnBin, 'copy',
            '-m', message,
            svn_url,
            '%s/tags/%s' % ( config['vcsRepo'], tag )
        ] )
    
    def vcsPush( self, config, refs ):
        pass
        
    def vcsGetRevision( self, config ) :
        svnBin = config['env']['svnBin']
        svn_info = self._callExternal( [ svnBin, 'info',  '--xml' ] )
        
        svn_info = xml.dom.minidom.parseString( svn_info )
        svn_info = svn_info.getElementsByTagName('commit')

        if len(svn_info):
            return svn_info[0].getAttribute('revision')
            
        return 'local'

    def vcsGetRefRevision( self, config, vcs_cache_dir, branch ) :
        svnBin = config['env']['svnBin']
        
        if branch == 'trunk':
            svn_repo_path = '{0}/{1}'.format( config['vcsRepo'], branch )
        else :
            svn_repo_path = '{0}/branches/{1}'.format( config['vcsRepo'], branch )
            
        res = self._callExternal( [ svnBin, 'info', svn_repo_path, '--xml' ] )
        
        res = xml.dom.minidom.parseString( res )
        self._rev = res.getElementsByTagName('commit')[0].getAttribute('revision')
        return self._rev

    def vcsListTags( self, config, vcs_cache_dir, tag_hint ) :
        svnBin = config['env']['svnBin']
        vcsRepo = config['vcsRepo']
        svn_repo_path = '{0}/tags/'.format( vcsRepo )
        res = self._callExternal( [ svnBin, 'ls', svn_repo_path ] ).strip().split("\n")
        return [v.replace(svn_repo_path, '').replace('/', '') for v in res ]

    def vcsExport( self, config, vcs_cache_dir, vcs_ref, dst_path ) :
        svnBin = config['env']['svnBin']
        
        if self._rev :
            if vcs_ref == 'trunk':
                svn_repo_path = '{0}/{1}@{2}'.format( config['vcsRepo'], vcs_ref, self._rev )
            else :
                svn_repo_path = '{0}/branches/{1}@{2}'.format( config['vcsRepo'], vcs_ref, self._rev )
        else :
            svn_repo_path = '{0}/tags/{1}'.format( config['vcsRepo'], vcs_ref )
        
        if os.path.isdir( vcs_cache_dir ):
            cnd = 'switch'
        else:
            cnd = 'checkout'
            
        self._callExternal( [ svnBin, cnd, svn_repo_path, vcs_cache_dir ] )
        self._callExternal( [ svnBin, 'export', vcs_cache_dir, dst_path ] )






from ..buildtool import BuildTool
from ..testtool import TestTool
from .sdkmantoolmixin import SdkmanToolMixIn

class mavenTool( SdkmanToolMixIn, BuildTool, TestTool ):
    """Apache Maven is a software project management and comprehension tool.

Home: https://maven.apache.org/

Auto-detected based on pom.xml

Expects clean, compile, package and test targets.
"""    
    def autoDetect( self, config ) :
        return self._autoDetectByCfg(
                config,
                [ 'pom.xml' ]
        )
    
    def _binName( self ):
        return 'mvn'

    def onPrepare( self, config ):
        self._callExternal( [ config['env']['mavenBin'], 'clean' ] )
    
    def onBuild( self, config ):
        self._callExternal( [ config['env']['mavenBin'], 'compile' ] )
        
    def onPackage( self, config ):
        self._callExternal( [ config['env']['mavenBin'], 'package' ] )

    def onCheck( self, config ):
        self._callExternal( [ config['env']['mavenBin'], 'test' ] )

            



from ..runenvtool import RunEnvTool

class gpgTool( RunEnvTool ):
    """The GNU Privacy Guard.

Home: https://www.gnupg.org/

gpgKeyServer is hkp://keyserver.ubuntu.com:80 by default
"""    
    def envNames( self ) :
        return ['gpgBin', 'gpgKeyServer']
   
    def _installTool( self, env ):
        self._requirePackages(['gnupg', 'gnupg2'])
        self._requireEmerge(['app-crypt/gnupg'])
        self._requirePacman(['gnupg'])
    
    def initEnv( self, env ):
        super(gpgTool, self).initEnv( env )
        env.setdefault('gpgKeyServer', 'hkp://keyserver.ubuntu.com:80')
        



import os, re

from ..buildtool import BuildTool

class releaseTool( BuildTool ):
    "FutoIn CID-specific release processing"
    
    RELEASE_FILE = os.path.join('futoin', 'cid', '__init__.py')
    CHANGELOG_FILE = 'CHANGELOG.txt'

    def autoDetect( self, config ) :
        return True

    def updateProjectConfig( self, config, updates ) :
        def py_updater( content ):
            if 'version' in updates:
                return re.sub(
                    r'^.*__version__.*$',
                    '__version__ = \'{0}\''.format(updates['version']),
                    content,
                    flags=re.MULTILINE
                )
        def cl_updater( content ):
            if 'version' in updates:
                return re.sub(
                    r'^=== \(next\) ===$',
                    '=== {0} ==='.format(updates['version']),
                    content,
                    count = 1,
                    flags=re.MULTILINE
                )
        res = self._updateTextFile(
            self.RELEASE_FILE,
            py_updater
        )
        res += self._updateTextFile(
            self.CHANGELOG_FILE,
            cl_updater
        )
        return res

    def initEnv( self, env ):
        self._have_tool = True




from ..runenvtool import RunEnvTool

class futoinTool( RunEnvTool ):
    """futoin.json updater as defined in FTN16.
    
Home: https://github.com/futoin/specs/blob/master/draft/ftn16_cid_tool.md

Auto-detected based on futoin.json

futoin.json is the only file read by FutoIn CID itself.
"""
    def autoDetect( self, config ) :
        return self._autoDetectByCfg( config, 'futoin.json' )

    def initEnv( self, env ):
        self._have_tool = True

    def updateProjectConfig( self, config, updates ) :
        def updater( json ):
            for f in ( 'name', 'version' ):
                if f in updates :
                        json[f] = updates[f]

        return self._updateJSONConfig( 'futoin.json', updater )




from ..buildtool import BuildTool
from .sdkmantoolmixin import SdkmanToolMixIn

class gradleTool( SdkmanToolMixIn, BuildTool ):
    """Gradle Build Tool.

Home: https://gradle.org/

Auto-detected based on build.gradle.
"""    
    def autoDetect( self, config ) :
        return self._autoDetectByCfg(
                config,
                [ 'build.gradle' ]
        )

    def onPrepare( self, config ):
        self._callExternal( [ config['env']['gradleBin'], '-q', '--no-daemon', 'clean' ] )
    
    def onBuild( self, config ):
        self._callExternal( [ config['env']['gradleBin'], '-q', '--no-daemon' ] )

    def onPackage( self, config ):
        self._callExternal( [ config['env']['gradleBin'], '-q', '--no-daemon', 'dists' ] )
            



import os

from ..buildtool import BuildTool

class gemTool( BuildTool ):
    """RubyGems: Find, install, and publish RubyGems.
    
Home: https://rubygems.org/

If rubyVer is equal to system then gems are installed in
user's folder gemDir.

gemDir is equal to ~/.gem by default.

gemInstallArgs is forcibly set by tool depending on its version.
"""    
    def getDeps( self ) :
        return [ 'ruby' ]
        
    def uninstallTool( self, env ):
        pass
    
    def envNames(self):
        return ['gemBin', 'gemDir', 'gemInstallArgs']
    
    def initEnv( self, env ) :
        installArgs = []
        
        if env['rubyVer'] == self.SYSTEM_VER:
            gemDir = os.path.join(os.environ['HOME'], '.gem')
            gemDir = env.setdefault('gemDir', gemDir)
            os.environ['GEM_HOME'] = gemDir
            
            self._addEnvPath( 'GEM_PATH', gemDir)
            self._addBinPath( os.path.join(gemDir, 'bin'))
            installArgs += ['--no-user-install', '--no-format-executable']

        super(gemTool, self).initEnv( env )

        if self._have_tool:
            version = self._callExternal([env['gemBin'], '--version'], verbose=False).strip()
            
            if version >= '2':
                installArgs += ['--no-document']
            else :
                installArgs += ['--no-ri', '--no-rdoc']
                
            env['gemInstallArgs'] = installArgs




from ..buildtool import BuildTool
from .gemtoolmixin import GemToolMixIn

class bundlerTool( GemToolMixIn, BuildTool ):
    """Bundler: The best way to manage a Ruby application's gems.
    
Home: http://bundler.io/

Auto-detected based on Gemfile.
"""
    def autoDetect( self, config ) :
        return self._autoDetectByCfg(
                config,
                [ 'Gemfile' ]
        )

    def initEnv( self, env ) :
        super(bundlerTool, self).initEnv( env, 'bundle' )

    def onPrepare( self, config ):
        self._callExternal( [ config['env']['bundlerBin'], 'install' ] )

    def onPackage( self, config ):
        self._callExternal( [ config['env']['bundlerBin'], 'install',
                             '--deployment', '--clean' ] )




import os

from ..runenvtool import RunEnvTool
from .bashtoolmixin import BashToolMixIn

class rvmTool( BashToolMixIn, RunEnvTool ):
    """Ruby Version Manager.
    
Home: https://rvm.io/

Stable RVM is used by default.
"""
    
    RVM_DIR_DEFAULT = os.path.join(os.environ['HOME'], '.rvm')
    RVM_LATEST = 'stable'
    RVM_GPG_KEY = '409B6B1796C275462A1703113804BB82D39DC0E3'
    RVM_GET = 'https://get.rvm.io'

    def getDeps( self ) :
        return [ 'bash', 'curl', 'gpg' ]

    def _installTool( self, env ):
        rvm_dir = env['rvmDir']
        rvm_get = env.get('rvmGet', self.RVM_GET) 
        rvm_gpg_key = env.get('rvmGpgKey', self.RVM_GPG_KEY)
        
        self._callExternal([
            env['gpgBin'], '--keyserver', env['gpgKeyServer'],
            '--recv-keys', rvm_gpg_key
        ], suppress_fail=True)

        os.environ['rvm_user_install_flag'] = '1'
        os.environ['rvm_auto_dotfiles_flag'] = '0'

        self._callBash( env,
              '{0} -sSL {1} | {2} -s {3} --path {4}'
               .format(env['curlBin'], rvm_get, env['bashBin'], env['rvmVer'], env['rvmDir']) )
            
    def updateTool( self, env ):
        self._callExternal([ env['rvmBin'], 'get', env['rvmVer'] ])
        
    def uninstallTool( self, env ):
        try:
            self._callExternal([ env['rvmBin'], 'implode', '--force' ])
        except:
            pass
        
        rvm_dir = env['rvmDir']

        if os.path.exists(rvm_dir):
            self._rmTree(rvm_dir)

        self._have_tool = False

    def envNames( self ) :
        return ['rvmVer', 'rvmDir', 'rvmBin', 'rvmGet', 'rvmGpgKey' ]

    def initEnv( self, env ) :
        for v in ['rvm_path', 'rvm_bin_path', 'rvm_prefix', 'rvm_version']:
            try: del os.environ[v]
            except: pass
        
        env.setdefault('rvmVer', self.RVM_LATEST) 
        rvm_dir = env.setdefault('rvmDir', self.RVM_DIR_DEFAULT)
        rvm_bin_dir = os.path.join(rvm_dir, 'bin')
        rvm_bin = env.setdefault('rvmBin', os.path.join(rvm_bin_dir, 'rvm'))
        self._have_tool = os.path.exists(rvm_bin)




from ..runenvtool import RunEnvTool

class zipTool( RunEnvTool ):
    """package and compress (archive) files.
"""    
    def _installTool( self, env ):
        self._requirePackages(['zip'])
        self._requireEmerge(['app-arch/zip'])
        self._requirePacman(['zip'])




from ..buildtool import BuildTool
from .sdkmantoolmixin import SdkmanToolMixIn

class antTool( SdkmanToolMixIn, BuildTool ):
    """Ant build tool for Java applications.
    
Home: http://ant.apache.org/

Ant is auto-detected based on build.xml.
The tool assumes the following targets: clean, compile, jar, run

Ant is setup through SDKMan!

Note: If detected Java version is less than 8 then Ant 1.9.8 is used.
"""
    def autoDetect( self, config ) :
        return self._autoDetectByCfg(
                config,
                [ 'build.xml' ]
        )
    
    def initEnv( self, env ) :
        if self._javaVersion( env ) < 8:
            env['antVer'] = '1.9.8'

        SdkmanToolMixIn.initEnv(self, env)

    def onPrepare( self, config ):
        self._callExternal( [ config['env']['antBin'], 'clean' ] )
    
    def onBuild( self, config ):
        self._callExternal( [ config['env']['antBin'], 'compile' ] )
        
    def onPackage( self, config ):
        self._callExternal( [ config['env']['antBin'], 'jar' ] )
        
    def onRun( self, config ):
        self._callExternal( [ config['env']['antBin'], 'run' ] )

            



import os

from ..buildtool import BuildTool
from .npmtoolmixin import NpmToolMixIn

class bowerTool( NpmToolMixIn, BuildTool ):
    """Bower: a package manager for the web.
    
Home: https://bower.io/

Auto-detected based on bower.json
"""
    BOWER_JSON = 'bower.json'
    
    def autoDetect( self, config ) :
        return self._autoDetectByCfg( config, self.BOWER_JSON )

    def loadConfig( self, config ) :
        content = self._loadJSONConfig( self.BOWER_JSON )
        if content is None: return
        
        f = 'name'
        if f in content and f not in config:
            config[f] = content[f]

    def updateProjectConfig( self, config, updates ) :
        def updater( json ):
            f = 'name'
            if f in updates :
                    json[f] = updates[f]
                    
            # version is deprecated
            if 'version' in json:
                del json['version']

        return self._updateJSONConfig( self.BOWER_JSON, updater )
    
    def onPrepare( self, config ):
        bowerBin = config['env']['bowerBin']
        self._callExternal( [ bowerBin, 'install' ] )

    def onPackage( self, config ):
        # Bower does not remove dev deps by itself
        if os.path.exists('bower_components'):
            self._rmTree('bower_components')
        
        bowerBin = config['env']['bowerBin']
        self._callExternal( [ bowerBin, 'install', '--production' ] )
    




import os, re

from .bashtoolmixin import BashToolMixIn

class SdkmanToolMixIn( BashToolMixIn ):
    def getDeps( self ) :
        return super( SdkmanToolMixIn, self ).getDeps() + [ 'sdkman', 'java' ]

    def _sdkName( self ):
        return self._name
    
    def _binName( self ):
        return self._name
    
    def _setAutoAnswer( self, env ):
        self._callBash( env,
            'grep -q "sdkman_auto_answer=true" {0} || echo "sdkman_auto_answer=true" >> {0}'
            .format(os.path.join(env['sdkmanDir'], 'etc', 'config')),
            verbose=False
        )
    
    def _callSdkman( self, env, cmd, verbose=True ):
        return self._callBash( env,
            'source {0} >/dev/null && sdk {1}'.format(
                env['sdkmanInit'],
                cmd
            ),
            verbose=verbose
        )
            
    def _javaVersion( self, env ):
        java_ver = self._callBash(env, '{0} -version 2>&1'.format(env['javaBin']), verbose=False)
        return int(re.search('version "1\.([0-9]+)\.', java_ver).group(1))
        
    def _installTool( self, env ):
        self._setAutoAnswer( env )
        self._callSdkman( env,
            'install {0} {1}'.format(
                self._sdkName(),
                env.get(self._name + 'Ver', '')
            )
        )
        
    def updateTool( self, env ):
        self._setAutoAnswer( env )
        self._callSdkman( env,
            'upgrade {0}'.format(
                self._sdkName()
            )
        )

        
    def uninstallTool( self, env ):
        tool_dir = os.path.join(env['sdkmanDir'], 'candidates', self._sdkName())
        if os.path.exists(tool_dir):
            print( 'Removing: {0}'.format(tool_dir) )
            self._rmTree(tool_dir)
        #self._setAutoAnswer( env )
        #self._callSdkman( env,
        #    'uninstall {0} {1}'.format(
        #        self._sdkName(),
        #        env.get(self._name + 'Ver', '')
        #    )
        #)

    def initEnv( self, env ) :
        tool_dir = os.path.join(env['sdkmanDir'], 'candidates', self._sdkName(), 'current')
        
        if not os.path.exists(tool_dir):
            return
        
        try:
            env_to_set = self._callSdkman( env,
                'use {0} {1} >/dev/null && env | grep -i {0}'
                .format(self._sdkName(), env.get(self._name + 'Ver', '')),
                verbose = False
            )
        except:
            return

        if env_to_set :
            self._updateEnvFromOutput(env_to_set)
            super(SdkmanToolMixIn, self).initEnv( env, self._binName() )



from ..runenvtool import RunEnvTool

class unzipTool( RunEnvTool ):
    """list, test and extract compressed files in a ZIP archive.
"""
    def _installTool( self, env ):
        self._requirePackages(['unzip'])
        self._requireEmerge(['app-arch/unzip'])
        self._requirePacman(['unzip'])




from ..runtimetool import RuntimeTool
from .bashtoolmixin import BashToolMixIn

class rubyTool( BashToolMixIn, RuntimeTool ):
    """Ruby is a dynamic, open source programming language.
    
Home: https://www.ruby-lang.org/en/

System Ruby is used by default.
If rubyVer is set then RVM is used to setup custom rubies.
That may lead to long time and resource consumption due to compilation,
if binary versions are not found for specific system.
"""
    def getDeps( self ) :
        return [ 'rvm' ]

    def _installTool( self, env ):
        ruby_ver = env['rubyVer']
        
        if ruby_ver == self.SYSTEM_VER:
            self._systemDeps()
            return
        
        self._buildDeps()
        self._callExternal([
            env['rvmBin'], 'install', ruby_ver, '--autolibs=read-only'
        ])
        self._callExternal([
            env['rvmBin'], 'cleanup', 'all'
        ])

    def updateTool( self, env ):
        if env['rubyVer'] != self.SYSTEM_VER:
            self._installTool( env )
        
    def uninstallTool( self, env ):
        ruby_ver = env['rubyVer']
        
        if ruby_ver != self.SYSTEM_VER:
            self._callExternal([
                env['rvmBin'], 'uninstall', env['rubyVer']
            ])
            self._have_tool = False

    def envNames( self ) :
        return ['rubyVer', 'rubyBin' ]

    def initEnv( self, env ) :
        ruby_ver = env.setdefault('rubyVer', self.SYSTEM_VER)

        rvm_dir = env['rvmDir']

        try:
            env_to_set = self._callBash( env,
                'source {0}/scripts/rvm && \
                rvm use {1} >/dev/null && \
                env | grep "rvm"'.format(rvm_dir, ruby_ver),
                verbose = False
            )
        except:
            return

        if env_to_set :
            self._updateEnvFromOutput(env_to_set)
            super(rubyTool, self).initEnv( env )
            
    def _buildDeps( self ):
        # APT
        #---
        self._requireDeb([
            'build-essential',
            'gawk',
            'libssl-dev',
            'make',
            'libc6-dev',
            'zlib1g-dev',
            'libyaml-dev',
            'libsqlite3-dev',
            'sqlite3',
            'autoconf',
            'libgmp-dev',
            'libgdbm-dev',
            'libncurses5-dev',
            'automake',
            'libtool',
            'bison',
            'pkg-config',
            'libffi-dev',
            'libgmp-dev',
            'libreadline-dev',
        ])
        

        # Extra repo before the rest
        #---
        self._requireYum(['epel-release'])

        self._requireRpm([
            'binutils',
            'patch',
            'libyaml-devel',
            'autoconf',
            'gcc',
            'gcc-c++',
            'glibc-devel',
            'readline-devel',
            'zlib-devel',
            'libffi-devel',
            'openssl-devel',
            'automake',
            'libtool',
            'bison',
            'sqlite-devel',
            'make',
            'm4',
            'gdbm-devel',
            'libopenssl-devel',
            'sqlite3-devel',
        ])
        
        self._requireEmergeDepsOnly(['dev-lang/ruby'])
        self._requirePacman(['ruby'])
        
    def _systemDeps( self ):
        self._requirePackages(['ruby'])
        self._requireEmerge(['dev-lang/ruby'])
        self._requirePacman(['ruby'])




class BashToolMixIn( object ):
    def getDeps( self ) :
        return [ 'bash' ]
    
    def _callBash( self, env, cmd, *args, **nargs ):
        return self._callExternal([
            env['bashBin'],
            '--noprofile', '--norc',
            '-c', cmd,
        ], *args, **nargs )




import os

from ..runenvtool import RunEnvTool
from .bashtoolmixin import BashToolMixIn

class phpbuildTool( BashToolMixIn, RunEnvTool ):
    """Builds PHP so that multiple versions can be used side by side.
    
Home: https://github.com/php-build/php-build
"""
    PHPBUILD_DIR_DEFAULT = os.path.join(os.environ['HOME'], '.phpbuild')
    PHPBUILD_LATEST = 'master'

    def getDeps( self ) :
        return [ 'bash', 'git' ]

    def _installTool( self, env ):
        phpbuild_dir = env['phpbuildDir']
        phpbuild_git = env.get('phpbuildGit', 'https://github.com/php-build/php-build.git')
        phpbuild_ver = env.get('phpbuildVer', self.PHPBUILD_LATEST) 
        
        self._callBash( env,
              'git clone {1} {0}; \
               cd {0} && git fetch && git reset --hard && git checkout {2}'
               .format(phpbuild_dir, phpbuild_git, phpbuild_ver ) )
            
    def updateTool( self, env ):
        phpbuild_dir = env['phpbuildDir']
        phpbuild_ver = env.get('phpbuildVer', self.PHPBUILD_LATEST) 

        self._callBash( env,
              'cd {0} && git fetch && git reset --hard && git checkout {1} && git pull --rebase'
               .format(phpbuild_dir, phpbuild_ver) )
    
    def uninstallTool( self, env ):
        phpbuild_dir = env['phpbuildDir']
        
        if os.path.exists(phpbuild_dir):
            self._rmTree(phpbuild_dir)
            
        self._have_tool = False
            
    def envNames( self ) :
        return ['phpbuildDir', 'phpbuildBin', 'phpbuildGit', 'phpbuildVer']
    
    def initEnv( self, env ) :
        phpbuild_dir = env.setdefault('phpbuildDir', self.PHPBUILD_DIR_DEFAULT)
        phpbuild_bin = env.setdefault('phpbuildBin', os.path.join(phpbuild_dir, 'bin', 'php-build'))
        self._have_tool = os.path.exists(phpbuild_bin)




import os

from ..runtimetool import RuntimeTool
from .bashtoolmixin import BashToolMixIn

class nodeTool( BashToolMixIn, RuntimeTool ):
    """Node.js is a JavaScript runtime built on Chrome's V8 JavaScript engine.
    
Home: https://nodejs.org/en/    
"""
    def getDeps( self ) :
        return [ 'nvm', 'bash' ]

    def _installTool( self, env ):
        nvm_dir = env['nvmDir']
        node_version = env['nodeVer']
        self._callBash( env,
                'source {0}/nvm.sh --no-use && nvm install {1}'
                .format(nvm_dir, node_version))
            
    def updateTool( self, env ):
        self._installTool( env )
        
    def uninstallTool( self, env ):
        nvm_dir = env['nvmDir']
        node_version = env['nodeVer']
        self._callBash( env,
              'source {0}/nvm.sh --no-use && nvm deactivate && nvm uninstall {1}'
              .format(nvm_dir, node_version) )
        self._have_tool = False

    def envNames( self ) :
        return ['nodeBin', 'nodeVer']
    
    def initEnv( self, env ) :
        node_version = env.setdefault('nodeVer', 'stable')

        nvm_dir = env['nvmDir']

        try:
            env_to_set = self._callBash( env,
                'source {0}/nvm.sh --no-use && \
                nvm use {1} >/dev/null && \
                env | egrep "(NVM_|\.nvm)"'.format(nvm_dir, node_version),
                verbose = False
            )
        except:
            return

        if env_to_set :
            self._updateEnvFromOutput(env_to_set)
            super(nodeTool, self).initEnv( env )




import os

from ..vcstool import VcsTool
from .bashtoolmixin import BashToolMixIn

class hgTool( BashToolMixIn, VcsTool ):
    """Mercurial SCM.
    
Home: https://www.mercurial-scm.org/

Auto-detected based on .hg folder
"""
    def getDeps( self ) :
        return [ 'bash' ]
    
    def _installTool( self, env ):
        self._requirePackages(['mercurial'])
        self._requireEmerge(['dev-vcs/mercurial'])
        self._requirePacman(['mercurial'])
    
    def autoDetect( self, config ) :
        return self._autoDetectVCS( config, '.hg' )

    def vcsGetRepo( self, config, wc_dir=None ):
        return self._callExternal( [
            config['env']['hgBin'], '--repository', wc_dir or config['wcDir'], 'paths', 'default'
        ] ).strip()
        
    def vcsCheckout( self, config, vcs_ref ):
        hgBin = config['env']['hgBin']
        wc_dir = config['wcDir']

        if os.path.isdir( '.hg' ):
            if 'vcsRepo' in config:
                remote_info = self.vcsGetRepo( config, '.' )
                if remote_info != config['vcsRepo'] :
                    raise RuntimeError( "Hg remote mismatch: " + remote_info )

            self._callExternal( [ hgBin, 'pull' ] )
        else :
            self._callExternal( [ hgBin, 'clone', config['vcsRepo'], wc_dir ] )

        self._callExternal( [ hgBin, 'checkout', '--check', vcs_ref ] )
    
    def vcsCommit( self, config, message, files ):
        hgBin = config['env']['hgBin']
        self._callExternal( [ hgBin, 'commit', '-m', message ] + files )
    
    def vcsTag( self, config, tag, message ):
        hgBin = config['env']['hgBin']
        self._callExternal( [ hgBin, 'tag', '-m', message, tag ] )
    
    def vcsPush( self, config, refs ):
        env = config['env']
        hgBin = env['hgBin']
        opts = []
        for r in refs :
            is_tag = self._callBash( env,
                '{0} tags | egrep "^{1}" || true'.format( hgBin, r )
            )
            
            if is_tag : continue
                
            opts.append( '-b' )
            opts.append( r )
        self._callExternal( [ hgBin, 'push' ] + opts )
        
    def vcsGetRevision( self, config ) :
        hgBin = config['env']['hgBin']
        return self._callExternal( [ hgBin, 'identify', '--id' ] ).strip()
    
    def _hgCache( self, config, vcs_cache_dir ):
        hgBin = config['env']['hgBin']

        if os.path.isdir( vcs_cache_dir ):
            remote_info = self.vcsGetRepo( config, vcs_cache_dir )

            if remote_info != config['vcsRepo'] :
                self._rmTree( vcs_cache_dir )
            else :
                self._callExternal( [ hgBin, '--cwd', vcs_cache_dir, 'pull' ] )

        if not os.path.isdir( vcs_cache_dir ):
            self._callExternal( [ hgBin, 'clone', config['vcsRepo'], vcs_cache_dir ] )

        
    def vcsGetRefRevision( self, config, vcs_cache_dir, branch ) :
        self._hgCache( config, vcs_cache_dir )
        hgBin = config['env']['hgBin']
        res = self._callExternal( [ hgBin, '--repository', vcs_cache_dir, 'branches' ] ).strip().split("\n")
        for r in res:
            r = r.split()
            
            if r[0] == branch:
                return r[1]
            
        raise RuntimeError( "Uknown Hg ref: {0}".format( branch ) )

    def vcsListTags( self, config, vcs_cache_dir, tag_hint ) :
        self._hgCache( config, vcs_cache_dir )
        hgBin = config['env']['hgBin']
        res = self._callExternal( [ hgBin, '--repository', vcs_cache_dir, 'tags' ] ).strip().split("\n")
        res = [ v.split()[0] for v in res ]
        del res[res.index('tip')]
        return res

    def vcsExport( self, config, vcs_cache_dir, vcs_ref, dst_path ) :
        self._callExternal([
            config['env']['hgBin'],
            '--repository', vcs_cache_dir,
            'archive',
            '--rev', vcs_ref,
            '--type', 'files',
            dst_path
        ])





from ..buildtool import BuildTool
from ..runenvtool import RunEnvTool

class dockerTool( BuildTool, RunEnvTool ):
    """Docker - Build, Ship, and Run Any App, Anywhere.
    
Home: https://www.docker.com/

Experimental support.
"""
    def autoDetect( self, config ) :
        return self._autoDetectByCfg(
                config,
                [ 'Dockerfile' ]
        )

    def getOrder( self ):
        return 10

    def onBuild( self, config ):
        self._callExternal( [ config['env']['dockerBin'], 'build' ] )
        




class NpmToolMixIn( object ):
    def getDeps( self ) :
        return [ 'node', 'npm' ]

    def _npmName( self ):
        return self._name
        
    def _installTool( self, env ):
        self._callExternal( [ env['npmBin'], 'install', '-g', self._npmName() ] )
        
    def updateTool( self, env ):
        self._callExternal( [ env['npmBin'], 'update', '-g', self._npmName() ] )
        
    def uninstallTool( self, env ):
        self._callExternal( [ env['npmBin'], 'uninstall', '-g', self._npmName() ] )
        self._have_tool = False




from ..runenvtool import RunEnvTool

class tarTool( RunEnvTool ):
    """The GNU version of the tar archiving utility.
"""    
    def _installTool( self, env ):
        self._requirePackages(['tar'])
        self._requireEmerge(['app-arch/tar'])
        self._requirePacman(['tar'])




import os

from ..buildtool import BuildTool

class npmTool( BuildTool ):
    """npm is the package manager for JavaScript.
    
Home: https://www.npmjs.com/    

Auto-detected based on package.json
"""
    PACKAGE_JSON = 'package.json'
    
    def autoDetect( self, config ) :
        return self._autoDetectByCfg( config, self.PACKAGE_JSON )
    
    def getDeps( self ) :
        return [ 'node' ]
        
    def updateTool( self, env ):
        self._callExternal( [ env['npmBin'], 'update', '-g', 'npm' ] )
        
    def uninstallTool( self, env ):
        pass
    
    def loadConfig( self, config ) :
        content = self._loadJSONConfig( self.PACKAGE_JSON )
        if content is None: return
        
        for f in ( 'name', 'version' ):
            if f in content and f not in config:
                config[f] = content[f]

    def updateProjectConfig( self, config, updates ) :
        def updater( json ):
            for f in ( 'name', 'version' ):
                if f in updates :
                        json[f] = updates[f]

        return self._updateJSONConfig( self.PACKAGE_JSON, updater )
    
    def onPrepare( self, config ):
        if os.path.exists( self.PACKAGE_JSON ):
            npmBin = config['env']['npmBin']
            self._callExternal( [ npmBin, 'install' ] )
        
    def onPackage( self, config ):
        if os.path.exists( self.PACKAGE_JSON ):
            npmBin = config['env']['npmBin']
            self._callExternal( [ npmBin, 'prune', '--production' ] )





from ..buildtool import BuildTool
from ..testtool import TestTool
from .sdkmantoolmixin import SdkmanToolMixIn

class sbtTool( SdkmanToolMixIn, BuildTool, TestTool ):
    """The interactive build tool (Scala).
    
Home: http://www.scala-sbt.org/

Auto-detected based on build.sbt

Installed via SDKMan!

First run of SBT may consume a lot of time on post-installation steps.
"""
    def autoDetect( self, config ) :
        return self._autoDetectByCfg(
                config,
                [ 'build.sbt' ]
        )
    
    def onPrepare( self, config ):
        self._callExternal( [ config['env']['sbtBin'], 'clean' ] )

    def onBuild( self, config ):
        self._callExternal( [ config['env']['sbtBin'], 'compile' ] )

    def onPackage( self, config ):
        self._callExternal( [ config['env']['sbtBin'], 'package' ] )
            
    def onRun( self, config ):
        self._callExternal( [ config['env']['sbtBin'], 'run' ] )
            
    def onCheck( self, config ):
        self._callExternal( [ config['env']['sbtBin'], 'test' ] )




import os, glob

from ..runtimetool import RuntimeTool

class javaTool( RuntimeTool ):
    """Java Runtime Environment.
    
Home: http://openjdk.java.net/

Due to issues with Oracle's licensing, cid
supports only automatic installation of OpenJDK.

javaVer supports only one digits like 7, 8, 9.
"""
    def envNames( self ) :
        return ['javaBin', 'javaVer']
    
    def _installTool( self, env ):
        if 'javaVer' in env:
            ver = env['javaVer'][0]
            self._requireDeb(['openjdk-{0}-jre-headless'.format(ver)])
            self._requireYum(['java-1.{0}.0-openjdk'.format(ver)])
            self._requireZypper(['java-1_{0}_0-openjdk'.format(ver)])
            self._requirePacman(['jre{0}-openjdk'.format(ver)])
            self._requireEmerge(['=dev-java/oracle-jre-bin-1.{0}*'.format(ver)])
        else :
            self._requireDeb(['default-jre-headless'])
            self._requireYum(['java-1.8.0-openjdk'])
            self._requireZypper(['java-1_8_0-openjdk'])
            self._requirePacman(['jre8-openjdk'])
            self._requireEmerge(['virtual/jre'])
    
    def uninstallTool( self, env ):
        pass

    def initEnv( self, env ) :
        if 'javaBin' in env or 'javaVer' not in env:
            super(javaTool, self).initEnv( env, 'java' )
            return
        
        ver = env['javaVer'][0]
        
        candidates = [
            # Debian / Ubuntu
            "/usr/lib/jvm/java-{0}-openjdk*/jre/bin/java".format(ver),
            # RedHat
            "/usr/lib/jvm/jre-1.{0}.0/bin/java".format(ver),
            # OpenSuse
            "/usr/lib*/jvm/jre-1.{0}.0/bin/java".format(ver),
            # Default oracle
            "/opt/jdk/jdk1.{0}*/bin/java".format(ver),
        ]
        
        for c in candidates:
            bin_name = glob.glob(c)
            
            if bin_name:
                env['javaBin'] = bin_name[0]
                self._addBinPath(os.path.basename(env['javaBin']), True)
                self._have_tool = True
                break
            




from __future__ import print_function, absolute_import

class PackageMixIn( object ):
    def _requireDeb(self, packages):
        apt_get = self._which('apt-get')
        
        if apt_get:
            self._trySudoCall(
                [apt_get, 'install', '-y'] + packages,
                errmsg = 'WARNING: you may need to install build deps manually !'
            )

    def _requireYum(self, packages):
        yum = self._which('dnf')
        
        if not yum:
            yum = self._which('yum')
        
        if yum:
            self._trySudoCall(
                [yum, 'install', '-y'] + packages,
                errmsg = 'WARNING: you may need to install build deps manually !'
            )

    def _requireZypper(self, packages):        
        zypper = self._which('zypper')

        if zypper:
            self._trySudoCall(
                [zypper, 'install', '-y'] + packages,
                errmsg='WARNING: you may need to install build deps manually !'
            )
            
    def _requireRpm(self, packages):
        self._requireYum(packages)
        self._requireZypper(packages)
    
    def _requirePackages(self, packages):
        self._requireDeb(packages)
        self._requireRpm(packages)
        
    def _requireEmerge(self, packages):
        emerge = self._which('emerge')

        if emerge:
            self._trySudoCall(
                [emerge] + packages,
                errmsg='WARNING: you may need to install build deps manually !'
            )

    def _requireEmergeDepsOnly(self, packages):
        self._requireEmerge(['--onlydeps'] + packages)
        
    def _requirePacman(self, packages):
        pacman = self._which('pacman')

        if pacman:
            self._trySudoCall(
                [pacman, '-S', '--noconfirm', '--needed'] + packages,
                errmsg='WARNING: you may need to install build deps manually !'
            )
        



import os

from ..buildtool import BuildTool

class cmakeTool( BuildTool ):
    """Build, Test and Package Your Software With CMake.
    
Home: https://cmake.org/

CMake creates a build folder and does all processing in it.
Build folder is configurable through cmakeBuildDir env.
"""    
    def getOrder( self ):
        return -10

    def autoDetect( self, config ) :
        return self._autoDetectByCfg( config, [
            'CMakeLists.txt'
        ] )
    
    def envNames( self ):
        return ['cmakeBin', 'cmakeBuildDir']
    
    def _installTool( self, env ):
        self._requireDeb(['build-essential'])
        self._requireRpm(['gcc', 'gcc-c++'])
        self._requirePacman(['gcc'])
            
        self._requirePackages(['cmake'])
        self._requireEmerge(['dev-util/cmake'])
        self._requirePacman(['cmake'])
        
    def initEnv( self, env ):
        env.setdefault('cmakeBuildDir', 'build')
        super(cmakeTool, self).initEnv(env)
        
    def onPrepare( self, config ):
        build_dir = config['env']['cmakeBuildDir']

        if os.path.exists(build_dir):
            self._rmTree(build_dir)
        
    def onBuild( self, config ):
        build_dir = config['env']['cmakeBuildDir']
        
        if os.path.exists(build_dir):
            self._callExternal( [ config['env']['cmakeBin'], build_dir ] )
        else:
            os.mkdir( build_dir )
            os.chdir( build_dir )
            self._callExternal( [ config['env']['cmakeBin'], config['wcDir'] ] )
            os.chdir( config['wcDir'] )
            
        self._callExternal( [ config['env']['cmakeBin'], '--build', build_dir ] )
        




from __future__ import print_function, absolute_import

import os, fnmatch, glob

from ..runtimetool import RuntimeTool
from .bashtoolmixin import BashToolMixIn

class phpTool( BashToolMixIn, RuntimeTool ):
    """PHP is a popular general-purpose scripting language that is especially suited to web development.
    
Home: http://php.net/

By default system PHP is used.

However, if phpVer is set then we use php-build which make consume a lot of time and
resources due to lack of trusted binary builds.
"""    
    PHP_DIR = os.path.join(os.environ['HOME'], '.php')
    
    def getDeps( self ) :
        return [ 'bash', 'phpbuild' ]
    
    def _installTool( self, env ):
        php_ver = env['phpVer']
        
        if php_ver == self.SYSTEM_VER:
            self._systemDeps()
            return

        php_dir = env['phpDir']
        
        try: os.makedirs(php_dir)
        except: pass

        self._buildDeps(env)

        old_tmpdir = os.environ.get('TMPDIR', '/tmp')
        os.environ['TMPDIR'] = os.path.join(php_dir, '..')
        self._callExternal( [ env['phpbuildBin'], php_ver, env['phpDir'] ] )
        os.environ['TMPDIR'] = old_tmpdir
    
    def updateTool( self, env ):
        pass
    
    def uninstallTool( self, env ):
        if env['phpVer'] == self.SYSTEM_VER:
            return super(phpTool, self).uninstallTool( env )

        php_dir = env['phpDir']

        if os.path.exists(php_dir):
            self._rmTree(php_dir)

        self._have_tool = False
    
    def envNames( self ) :
        return ['phpDir', 'phpBin', 'phpVer', 'phpfpmBin']
    
    def initEnv( self, env ) :
        php_ver = env.setdefault('phpVer', self.SYSTEM_VER)
        
        if php_ver == self.SYSTEM_VER:
            super(phpTool, self).initEnv( env )
            if not self._have_tool: return

            env.setdefault('phpDir', '/usr')
            
            php_fpm = glob.glob(os.path.join(env['phpDir'], 'sbin', 'php*-fpm*'))
            if php_fpm:
                env.setdefault('phpfpmBin', php_fpm[0])
            return
        else:
            def_dir = os.path.join(env['phpbuildDir'], 'share', 'php-build', 'definitions')
            
            if not os.path.exists(def_dir):
                return
            
            defs = os.listdir(def_dir)
            defs = fnmatch.filter(defs, php_ver + '*')
            
            if not defs:
                raise RuntimeError('PHP version %s not found' % php_ver)
            
            def castver(v):
                try: return int(v)
                except: return -1
            
            defs.sort(key=lambda v: [castver(u) for u in v.split('.')])
            php_ver = defs[-1]
            
            env['phpVer'] = php_ver
        
        php_dir = env.setdefault('phpDir', os.path.join(self.PHP_DIR, php_ver))
        php_bin_dir = os.path.join(php_dir, 'bin')
        php_bin = os.path.join(php_bin_dir, 'php')
        
        if os.path.exists(php_bin):
            self._have_tool = True
            self._addBinPath( php_bin_dir, True )
            env.setdefault('phpBin', php_bin)
            env.setdefault('phpfpmBin', os.path.join(php_dir, 'sbin', 'php-fpm') )
            
    def _buildDeps( self, env ):
        # APT
        #---
        self._requireDeb([
            'build-essential',
            'bison',
            'automake',
            'autoconf',
            'libtool',
            're2c',
            'libcurl4-openssl-dev',
            'libtidy-dev',
            'libpng-dev',
            'libmcrypt-dev',
            'libjpeg-dev',
            'libreadline-dev',
            'libbz2-dev',
            'libc-client-dev',
            'libdb-dev',
            'libedit-dev',
            'libenchant-dev',
            'libevent-dev',
            'libexpat1-dev',
            'libfreetype6-dev',
            'libgcrypt11-dev',
            'libgd2-dev',
            'libglib2.0-dev',
            'libgmp3-dev',
            'libicu-dev',
            'libjpeg-dev',
            'libkrb5-dev',
            'libldap2-dev',
            'libmagic-dev',
            'libmhash-dev',
            'libmysqlclient-dev',
            'libonig-dev',
            'libpam0g-dev',
            'libpcre3-dev',
            'libpng-dev',
            'libpq-dev',
            'libpspell-dev',
            'libqdbm-dev',
            'librecode-dev',
            'libsasl2-dev',
            'libsnmp-dev',
            'libsqlite3-dev',
            'libssl-dev',
            'libwrap0-dev',
            'libxmltok1-dev',
            'libxml2-dev',
            'libvpx-dev',
            'libxslt1-dev',
            'unixodbc-dev',
            'zlib1g-dev',
        ])

        # Extra repo before the rest
        #---
        self._requireYum(['epel-release'])
        
        self._requireRpm([
            'binutils',
            'patch',
            'git',
            'gcc',
            'gcc-c++',
            'make',
            'autoconf',
            'automake',
            'libtool',
            'bison',
            're2c',
            'glibc-devel',
            'libxml2-devel',
            'pkgconfig',
            'openssl-devel',
            'curl-devel',
            'libpng-devel',
            'libjpeg-devel',
            'libXpm-devel',
            'freetype-devel',
            'gmp-devel',
            'libmcrypt-devel',
            'aspell-devel',
            'recode-devel',
            'libicu-devel',
            'oniguruma-devel',
            'libtidy-devel',
            'libxslt-devel',
            'readline-devel',
            'zlib-devel',
            'pcre-devel',
        ])
        
        self._requireYum([
            'bzip2-devel',
            'mysql-devel',
        ])

        self._requireZypper([
            'libbz2-devel',
            'libmysqlclient-devel',
        ])
        
        self._requireEmergeDepsOnly(['dev-lang/php'])
        self._requirePacman([
            'patch',
            'git',
            'gcc',
            'make',
            'autoconf',
            'automake',
            'libtool',
            'bison',
            're2c',
            'glibc',
            'libxml2',
            'openssl',
            'curl',
            'libpng',
            'libjpeg',
            'libxpm',
            'freetype2',
            'gmp',
            'libmcrypt',
            'aspell',
            'recode',
            'icu',
            'oniguruma',
            'tidy',
            'libxslt',
            'readline',
            'zlib',
            'pcre',
        ])
        
        #---
        systemctl = self._which('systemctl')

        if systemctl:
            self._requireDeb(['libsystemd-dev'])
            self._requireRpm(['systemd-devel'])
            with_systemd = ' --with-fpm-systemd'
        else:
            with_systemd = ' --without-fpm-systemd'
            
        multiarch = None
        dpkgarch = self._which('dpkg-architecture')

        if dpkgarch :
            multiarch = self._callExternal([dpkgarch, '-qDEB_HOST_MULTIARCH']).strip()

        if multiarch :
            if os.path.exists(os.path.join('/usr/include', multiarch, 'curl')):
                curl_dir = os.path.join(env['phpDir'], '..', 'curl')

                try:
                    os.mkdir(curl_dir)
                    os.symlink(os.path.join('/usr/include', multiarch), os.path.join(curl_dir, 'include'))
                    os.symlink(os.path.join('/usr/lib', multiarch), os.path.join(curl_dir, 'lib'))
                except Exception as e:
                    #print(e)
                    pass
            else:
                curl_dir = '/usr/include'

            with_libdir = ' --with-libdir={0} --with-curl={1}'.format(
                os.path.join('lib', multiarch),
                curl_dir,
            )
        else:
            with_libdir = ''
        #---
        cpu_count = int(self._callBash( env, 'cat /proc/cpuinfo | grep -c vendor' ))
        
        if cpu_count <= 0:
            cpu_count = 1
        
        os.environ['PHP_BUILD_EXTRA_MAKE_ARGUMENTS'] = '-j{0}'.format(cpu_count)
       
        os.environ['PHP_BUILD_CONFIGURE_OPTS'] = ' \
            --disable-debug \
            --with-regex=php \
            --enable-calendar \
            --enable-sysvsem \
            --enable-sysvshm \
            --enable-sysvmsg \
            --enable-bcmath \
            --disable-cgi \
            --disable-phpdbg \
            --enable-fpm \
            --with-bz2 \
            --enable-ctype \
            --without-db4 \
            --without-qdbm \
            --without-gdbm \
            --with-iconv \
            --enable-exif \
            --enable-ftp \
            --with-gettext \
            --enable-mbstring \
            --with-onig=/usr \
            --with-pcre-regex=/usr \
            --enable-shmop \
            --enable-sockets \
            --enable-wddx \
            --with-libxml-dir=/usr \
            --with-zlib \
            --with-kerberos=/usr \
            --with-openssl=/usr \
            --enable-soap \
            --enable-zip \
            --with-mhash=yes \
            --with-system-tzdata \
            ' + with_systemd + with_libdir

    def _systemDeps( self ):
        self._requireDeb([
            'php.*-cli',
            'php.*-fpm',
            "php.*-apcu",
            "php.*-curl",
            "php.*-gd",
            "php.*-geoip",
            "php.*-gmp",
            "php.*-imagick",
            "php.*-imap",
            "php.*-intl",
            "php.*-json",
            "php.*-ldap",
            "php.*-mcrypt",
            "php.*-msgpack",
            "php.*-ssh2",
            "php.*-soap",
            "php.*-xml",
            "php.*-xmlrpc",
            "php.*-xsl",
        ])
        
        # SuSe-like
        self._requireZypper([
            'php?',
            'php*-fpm',
            'php*-bcmath',
            'php*-bz2',
            'php*-calendar',
            'php*-ctype',
            'php*-curl',
            'php*-dom',
            'php*-exif',
            'php*-fileinfo',
            'php*-gettext',
            'php*-gmp',
            'php*-iconv',
            'php*-imap',
            'php*-intl',
            'php*-json',
            'php*-ldap',
            'php*-mbstring',
            'php*-mcrypt',
            'php*-pcntl',
            'php*-pdo',
            'php*-phar',
            'php*-soap',
            'php*-sockets',
            'php*-tidy',
            'php*-xmlreader',
            'php*-xmlrpc',
            'php*-xmlwriter',
            'php*-xsl',
            'php*-zip',
            'php*-zlib',
        ])
            
        # RedHat-like
        self._requireYum([
            'php-cli',
            'php-fpm',
            'php-pecl-apcu',
            'php-pecl-imagick',
            'php-pecl-msgpack',
            'php-pecl-ssh2',
            'php-pecl-zendopcache',
        ])
        
        try:
            self._requireDeb([
                "php.*-mbstring",
                "php.*-opcache",
                "php.*-zip",
            ])
        except:
            pass

        self._requireEmerge(['dev-lang/php'])
        self._requirePacman(['php'])
        



class PipToolMixIn( object ):
    def getDeps( self ) :
        return [ 'pip' ]

    def _pipName( self ):
        return self._name

    def _installTool( self, env ):
        self._callExternal( [ env['pipBin'], 'install', '-q', self._pipName() ] )
        
    def updateTool( self, env ):
        self._callExternal( [ env['pipBin'], 'install', '-q', '--upgrade', self._pipName() ] )
        
    def uninstallTool( self, env ):
        self._callExternal( [ env['pipBin'], 'uninstall', '-q', self._pipName() ] )
        self._have_tool = False

    def _requirePythonDev( self, env ):
        if int(env['pythonVer'].split('.')[0]) == 3:
            self._requireDeb(['python3-dev'])
            self._requireZypper(['python3-devel'])
            self._requireYum(['epel-release'])
            self._requireYum(['python34-devel'])
            self._requirePacman(['python'])
        else:
            self._requireDeb(['python-dev'])
            self._requireZypper(['python-devel'])
            self._requireYum(['epel-release'])
            self._requireYum(['python-devel'])
            self._requirePacman(['python2'])
            
        self._requireEmergeDepsOnly(['dev-lang/python'])
        
    def _requirePip( self, env, package ):
        self._callExternal( [ env['pipBin'], 'install', '-q', package ] )



from __future__ import print_function, absolute_import

import os
import sys
import subprocess
import importlib
import json
import datetime
import re
import gzip
import shutil
import stat
import time
import fnmatch
import fcntl
from collections import OrderedDict

from .mixins.path import PathMixIn
from .mixins.util import UtilMixIn

from .vcstool import VcsTool
from .rmstool import RmsTool
from .buildtool import BuildTool
from .testtool import TestTool
from .migrationtool import MigrationTool

__all__ = ['CIDTool']

def _call_cmd( cmd ):
    print( 'Call: ' + subprocess.list2cmdline( cmd ), file=sys.stderr )
    subprocess.call( cmd, stdin=subprocess.PIPE )    

def cid_action( f ):
    def custom_f( self, *args, **kwargs ) :
        config = self._config

        try:
            fn = f.func_name
        except AttributeError:
            fn = f.__name__

        if fn in config.get('actions', {}) :
            for a in config['actions'][fn] :
                if a == '<default>':
                    f( self, *args, **kwargs )
                else :
                    _call_cmd( ['sh', '-c', a] )
        else :
            f( self, *args, **kwargs )
    return custom_f

class CIDTool( PathMixIn, UtilMixIn ) :
    TO_GZIP = '\.(js|json|css|svg|txt)$'
    VCS_CACHE_DIR = 'vcs'
    
    DEPLOY_LOCK_FILE = '.futoin.lock'
    _deploy_lock = False

    def __init__( self, overrides ) :
        self._startup_env = dict(os.environ)
        self._tool_impl = None
        self._overrides = overrides
        self._initConfig()
        
    def _forEachTool( self, cb, allow_failure=False, base=None ) :
        config = self._config
        tool_impl = self._tool_impl
        tools = config['tools']

        for t in tools :
            t = tool_impl[t]

            if base and not isinstance(t, base):
                continue

            try:
                cb( config, t )
            except RuntimeError as e:
                if not allow_failure:
                    raise e
            
    def _getVcsTool( self ):
        config = self._config
        vcstool = config.get('vcs', None)
        
        if not vcstool:
            print( 'Unknown VCS. Please set through --vcsRepo or project manifest' )
            sys.exit( 1 )
            
        vcstool = self._tool_impl[vcstool]
            
        if not config.get('vcsRepo', None): # also check it set
            config['vcsRepo'] = vcstool.vcsGetRepo( config )
            
            if not config['vcsRepo']:
                print( 'Unknown VCS repo. Please set through --vcsRepo or project manifest' )
                sys.exit( 1 )
        
        return vcstool

    def _getRmsTool( self ):
        config = self._config
        rmstool = config.get('rms', None)
        
        if not rmstool:
            print( 'Unknown RMS. Please set through --rmsRepo or project manifest' )
            sys.exit( 1 )
        
        if not config.get('rmsRepo', None): # also check it set
            print( 'Unknown RMS repo. Please set through --rmsRepo or project manifest' )
            sys.exit( 1 )
        
        return self._tool_impl[rmstool]

    def _processWcDir( self ):
        config = self._config
        wcDir = config['wcDir']
        
        if not os.path.exists(wcDir):
            os.makedirs(wcDir)
        
        if wcDir != os.getcwd():
            os.chdir( wcDir )
            self._overrides['wcDir'] = config['wcDir'] = '.'
            self._initConfig()

    @cid_action
    def tag( self, branch, next_version=None ):
        if next_version and not re.match('^[0-9]+\.[0-9]+\.[0-9]+$', next_version):
            print( 'Valid version format: x.y.z', file=sys.stderr )
            sys.exit( 1 )
            
        self._processWcDir()
            
        config = self._config
        vcstool = self._getVcsTool()
        
        # make a clean checkout
        vcstool.vcsCheckout( config, branch )
        self._initConfig()
        config = self._config
        
        # Set new version
        if next_version is None:
            if 'version' in config :
                next_version = config['version']
            else :
                raise RuntimeError( 'current version is unknown' )

            next_version = next_version.split('.')
            next_version[-1] = str(int(next_version[-1]) + 1)
            next_version = '.'.join( next_version )
        config['version'] = next_version
        
        # Update files for release
        to_commit = []        
        self._forEachTool(
            lambda config, t: to_commit.extend(
                t.updateProjectConfig( config, { 'version' : next_version } )
            )
        )
        
        # Commit updated files
        message = "Updated for release %s %s" % ( config['name'], config['version'] )
        vcstool.vcsCommit( config, message, to_commit )
        
        # Craete a tag
        tag = "v%s" % next_version
        message = "Release %s %s" % ( config['name'], config['version'] )
        vcstool.vcsTag( config, tag, message )

        # Push changes for DVCS
        vcstool.vcsPush( config, [ branch, tag ] )

    @cid_action
    def prepare( self, vcs_ref ):
        self._processWcDir()
        
        config = self._config

        # make a clean checkout
        if vcs_ref:
            vcstool = self._getVcsTool()

            vcstool.vcsCheckout( config, vcs_ref )
            self._initConfig()

        #--
        self._forEachTool(
            lambda config, t: t.onPrepare( config ),
            base=BuildTool
        )

    @cid_action
    def build( self ):
        self._processWcDir()
        
        self._forEachTool(
            lambda config, t: t.onBuild( config ),
            base=BuildTool
        )

    @cid_action
    def package( self ):
        self._processWcDir()
        
        self._forEachTool(
            lambda config, t: t.onPackage( config ),
            base=BuildTool
        )
        
        #---
        config = self._config
        package_file = config.get( 'package_file', None )

        if package_file:
            self._lastPackage = package_file
            return

        #---
        package_content = config.get( 'package', [ '.' ] )
        if type(package_content) != type([]):
            package_content = [ package_content ] 
        package_content.sort()
        package_content_cmd = subprocess.list2cmdline( package_content )
        
        # Note: It is assumed that web root is in the package content
        #---
        walk_list = os.walk( config.get( 'webcfg', {}).get( 'root', '.' ) )
        to_gzip_re = re.compile( self.TO_GZIP, re.I )
        for ( path, dirs, files ) in walk_list :
            for f in files :
                if to_gzip_re.search( f ):
                    f = os.path.join( path, f )
                    with open(f, 'rb') as f_in:
                        with gzip.open(f + '.gz', 'wb', 9) as f_out:
                            shutil.copyfileobj(f_in, f_out)
        
        #---
        checksums_file = '.package.checksums'
        try:
            package_content.index( '.' )
        except ValueError:
            package_content.append( checksums_file )
        cmd = 'find {0} -type f | sort | xargs sha512sum >{1}'.format(
                package_content_cmd, checksums_file )
        _call_cmd( ['sh', '-c', cmd] )
        
        #---
        buildTimestamp = datetime.datetime.utcnow().strftime( '%Y%m%d%H%M%S' )
        name = config['name'].split('/')[-1]
        version = config['version']
        vcs_ref = config.get( 'vcsRef', None )
        
        # Note: unless run in clean ci_build process,
        # all builds must be treated as snapshots/CI builds
        if vcs_ref == 'v' + version :
            package_file = '{0}-{1}-{2}'.format(
                    name, version, buildTimestamp )
        else :
            vcstool = self._getVcsTool()
            vcs_ref = vcstool.vcsGetRevision( config )
            package_file = '{0}-CI-{1}-{2}-{3}'.format(
                    name, version, vcs_ref, buildTimestamp )

        if 'target' in config:
            package_file += '-{0}'.format( config['target'] )
            
        package_file += '.txz'
        _call_cmd( ['tar', 'cJf', package_file,
                    '--exclude=' + package_file, '--exclude-vcs' ] + package_content )
        self._lastPackage = package_file
    
    @cid_action
    def check( self ):
        self._processWcDir()
        
        self._forEachTool(
            lambda config, t: t.onCheck( config ),
            allow_failure = self._config.get('permissiveChecks', False),
            base=TestTool
        )

    @cid_action
    def promote( self, package, rms_pool ):
        config = self._config
        rmstool = self._getRmsTool()
        rmstool.rmsPromote( config, package, rms_pool )
        
    @cid_action
    def migrate( self, location ):
        self._processWcDir()
        
        self._forEachTool(
            lambda config, t: t.onMigrate( config, location ),
            base=MigrationTool
        )
        
    def _deployLock( self ):
        self._deploy_lock = open(self.DEPLOY_LOCK_FILE, 'w')
        try:
            fcntl.flock(self._deploy_lock, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except:
            print('FAILED to acquire deploy lock! ', file=sys.stderr)
            sys.exit( 1 )
    
    def _deployUnlock( self ):
        fcntl.flock(self._deploy_lock, fcntl.LOCK_UN)
        self._deploy_lock.close()
        self._deploy_lock = None
    
    @cid_action
    def deploy( self, mode, p1, p2=None ):
        # Get to deploy folder
        deploy_dir = self._config['deployDir']
        if deploy_dir:
            if not os.path.exists( deploy_dir ) :
                os.makedirs( deploy_dir )
            os.chdir( deploy_dir )

        self._deployLock()
        
        if mode == 'rms':
            self._rms_deploy(p1, p2)
        elif mode == 'vcsref':
            self._vcsref_deploy(p1)
        elif mode == 'vcstag':
            self._vcstag_deploy(p1)
        else:
            raise RuntimeError( 'Not supported deploy mode: ' + mode )
        
    def _rms_deploy( self, rms_pool, package=None ):
        config = self._config
        rmstool = self._getRmsTool()

        # Find out package to deploy
        package_list = rmstool.rmsGetList( config, rms_pool, package )

        if package:
            package_list = fnmatch.filter(package_list, package)
            
        if not package_list:
            print( "No package found", file = sys.stderr )
            sys.exit( 1 )
            
        package = self._getLatest(package_list)
            
        # cleanup first, in case of incomplete actions
        self._deployCleanup( [package] )

        # Prepare package name components
        package_basename = os.path.basename( package )
        ( package_noext, package_ext ) = os.path.splitext( package_basename )
        
        # Check if already deployed:
        if os.path.exists( package_noext ) and not config['reDeploy']:
            print( "Package has been already deployed. Use --redeploy.")
            return
        
        # Retrieve package, if not available
        if not os.path.exists( package_basename ) :
            rmstool.rmsRetrieve( config, rms_pool, package )
            
        package_noext_tmp = package_noext + '.tmp'
        
        # Prepare temporary folder
        os.mkdir( package_noext_tmp )
        
        # Unpack package to temporary folder
        if package_ext == '.txz':
            _call_cmd( ['tar', 'xJf', package_basename, '-C', package_noext_tmp ] )
        elif package_ext == '.tbz2':
            _call_cmd( ['tar', 'xjf', package_basename, '-C', package_noext_tmp ] )
        elif package_ext == '.tgz':
            _call_cmd( ['tar', 'xzf', package_basename, '-C', package_noext_tmp ] )
        elif package_ext == '.tar':
            _call_cmd( ['tar', 'xf', package_basename, '-C', package_noext_tmp ] )
        else:
            raise RuntimeError( 'Not supported package format: ' + package_ext )
        
        # Common processing
        self._deployCommon( package_noext_tmp, package_noext, [package] )
        
    def _vcsref_deploy( self, vcs_ref ):
        config = self._config
        vcstool = self._getVcsTool()

        # Find out package to deploy
        vcs_cache = self.VCS_CACHE_DIR
        rev = vcstool.vcsGetRefRevision( config, vcs_cache, vcs_ref )
            
        if not rev:
            print( "No VCS refs found", file = sys.stderr )
            sys.exit( 1 )
            
        target_dir = vcs_ref.replace(os.sep, '_').replace(':', '_')
        target_dir += '__' + rev
            
        # cleanup first, in case of incomplete actions
        self._deployCleanup( [vcs_cache, target_dir] )
        
        # Check if already deployed:
        if os.path.exists( target_dir ) and not config['reDeploy']:
            print( "Package has been already deployed. Use --redeploy.")
            return
           
        # Retrieve tag
        target_tmp = target_dir + '.tmp'
        vcstool.vcsExport( config, vcs_cache, vcs_ref, target_tmp )

        # Common processing
        self._deployCommon( target_tmp, target_dir, [vcs_cache] )      

    def _vcstag_deploy( self, vcs_ref ):
        config = self._config
        vcstool = self._getVcsTool()

        # Find out package to deploy
        vcs_cache = self.VCS_CACHE_DIR
        tag_list = vcstool.vcsListTags( config, vcs_cache, vcs_ref )

        if vcs_ref:
            tag_list = fnmatch.filter(tag_list, vcs_ref)
            
        if not tag_list:
            print( "No tags found", file = sys.stderr )
            sys.exit( 1 )
            
        vcs_ref = self._getLatest(tag_list)
        target_dir = vcs_ref.replace(os.sep, '_').replace(':', '_')
            
        # cleanup first, in case of incomplete actions
        self._deployCleanup( [vcs_cache, target_dir] )
        
        # Check if already deployed:
        if os.path.exists( target_dir ) and not config['reDeploy']:
            print( "Package has been already deployed. Use --redeploy.")
            return
           
        # Retrieve tag
        vcs_ref_tmp = target_dir + '.tmp'
        vcstool.vcsExport( config, vcs_cache, vcs_ref, vcs_ref_tmp )

        # Common processing
        self._deployCommon( vcs_ref_tmp, target_dir, [vcs_cache] )        
        
    def _getLatest( self, verioned_list ):
        def castver(v):
            res = re.split(r'[\W_]+', v)
            for (i, vc) in enumerate(res):
                try: res[i] = int(vc, 10)
                except: pass
            return res
            
        verioned_list.sort(key=castver)
        return verioned_list[-1]

    def _deployCommon( self, tmp, dst, cleanup_whitelist ):
        config = self._config
        
        # Setup persistent folders
        persistent_dir = os.path.abspath( config['env'].get('persistentDir', 'persistent') )
        wdir_wperm = stat.S_IRUSR | stat.S_IXUSR | \
                    stat.S_IRGRP | stat.S_IXGRP | \
                    stat.S_IWUSR | stat.S_IWGRP
        
        for d in config.get('persistent', []) :
            pd = os.path.join( persistent_dir, d )
            dd = os.path.join( tmp, d )

            if not os.path.isdir( pd ) :
                os.makedirs( pd, wdir_wperm )
                
            if os.path.exists( dd ):
                shutil.copytree( dd, pd )
                self._rmTree( dd )
            
            os.symlink( pd, dd )
            
        # Setup read-only permissions
        dir_perm = stat.S_IRUSR | stat.S_IXUSR | stat.S_IRGRP | stat.S_IXGRP
        file_perm = stat.S_IRUSR | stat.S_IRGRP
        walk_list = os.walk( tmp )
        os.chmod( tmp, dir_perm )
        
        for ( path, dirs, files ) in walk_list :
            for d in dirs :
                os.chmod( os.path.join( path, d ), dir_perm )
            for f in files :
                os.chmod( os.path.join( path, f ), file_perm )
                
        # Build
        if config.get('deployBuild', False):
            cwd = os.getcwd()
            config['wcDir'] = tmp
            self.prepare(None)
            self.build()
            os.chdir(cwd)
            
        # Complete migration
        self.migrate( tmp )
        
        # Setup per-user services
        self._deployServices( tmp )
        
        # Move in place
        if os.path.exists( dst ):
            # re-deploy case
            os.rename( dst, dst + '.tmprm' )
        os.rename( tmp, dst )
        os.symlink( dst, 'current.tmp' )
        os.rename( 'current.tmp', 'current' )
        
        # Re-run
        self.run( 'reload' )
        
        # Cleanup old packages and deploy dirs
        self._deployCleanup( cleanup_whitelist )
        self._deployUnlock()
        
    def _deployCleanup( self, whitelist ):
        if os.path.exists('current'):
            whitelist.append( os.path.basename(os.readlink('current')) )
            
        whitelist += ['current', 'persistent', self.DEPLOY_LOCK_FILE]

        for f in os.listdir( '.' ):
            ( f_noext, f_ext ) = os.path.splitext( f )
            
            if f in whitelist:
                continue
            
            if os.path.isdir(f):
                self._rmTree(f)
            else:
                os.chmod( f, stat.S_IRWXU )
                os.remove( f )
            
            
    def _deployServices( self, subdir ):
        pass
    
    @cid_action
    def run( self, command ):
        self._processWcDir()
        
        config = self._config
        if config.get('vcs', None) :
            self.runDev( command )
            return

    def runDev( self, command ):
        pass
   
    @cid_action
    def ci_build( self, vcs_ref, rms_pool ):
        config = self._config
        wcDir = config['wcDir']

        if os.path.exists( wcDir ) and wcDir != os.getcwd():
            os.rename( wcDir, '{0}.bak{1}'.format(wcDir, int(time.time())) )
        
        self._lastPackage = None
        self.prepare( vcs_ref )
        self.build()
        self.package()
        self.check()
        self.promote( self._lastPackage, rms_pool )

    def tool_exec( self, tool, args ):
        t = self._tool_impl[tool]
        bin = self._config['env'].get(tool + 'Bin')

        if bin :
            _call_cmd([bin] + args)
        else :
            raise NotImplementedError( "Tool exec has not been implemented for %s" % tool )
    
    def tool_install( self, tool ):
        config = self._config
        env = config['env']
        
        if self._isExternalToolsSetup( env ):
            raise RuntimeError( "Tools must be installed externally (env config)" )

        if tool :
            tools = [tool]
        else :
            tools = config['tools']

        for tool in tools:
            t = self._tool_impl[tool]
            t.requireInstalled( env )

    def tool_uninstall( self, tool ):
        config = self._config
        env = config['env']
        
        if self._isExternalToolsSetup( env ):
            raise RuntimeError( "Tools must be uninstalled externally (env config)" )

        if tool :
            tools = [tool]
        else :
            tools = reversed(config['tools'])

        for tool in tools:
            t = self._tool_impl[tool]
            if t.isInstalled( env ):
                t.uninstallTool( env )

    def tool_update( self, tool ):
        config = self._config
        env = config['env']
        
        if self._isExternalToolsSetup( env ):
            raise RuntimeError( "Tools must be updated externally (env config)" )

        if tool :
            tools = [tool]
        else :
            tools = config['tools']

        for tool in tools:
            t = self._tool_impl[tool]
            t.updateTool( env )

    def tool_test( self, tool ):
        config = self._config
        env = config['env']

        if tool :
            tools = [tool]
        else :
            tools = config['tools']

        for tool in tools:
            t = self._tool_impl[tool]
            
            if not t.isInstalled( env ) :
                print( "Tool '%s' is missing" % tool )
                sys.exit( 1 )
                
    def tool_env( self, tool ):
        config = self._config
        env = config['env']

        if tool :
            tools = [tool]
        else :
            tools = config['tools']
            
        res = dict(os.environ)
        
        # remove unchanged vars
        for k, v in self._startup_env.items():
            if res.get(k, None) == v:
                del res[k]

        for tool in tools:
            self._tool_impl[tool].exportEnv(env, res)
            
        for k, v in res.items():
            print("{0}='{1}'".format(k, v.replace("'", "\\'").replace('\\', '\\\\')))

    def _tool_cmd( self, tool, base, method ):
        config = self._config
        env = config['env']

        t = self._tool_impl[tool]

        if isinstance(t, base):
            t.loadConfig( config ) # see self._initTools()
            getattr( t, method )( config )
        else:
            raise RuntimeError( '{0} tool does not support {1}'.format(tool, method) )
        
    def tool_prepare( self, tool ):
        self._tool_cmd( tool, BuildTool, 'onPrepare' )

    def tool_build( self, tool ):
        self._tool_cmd( tool, BuildTool, 'onBuild' )

    def tool_check( self, tool ):
        self._tool_cmd( tool, TestTool, 'onCheck' )

    def tool_package( self, tool ):
        self._tool_cmd( tool, BuildTool, 'onPackage' )

    def tool_migrate( self, tool ):
        self._tool_cmd( tool, MigrationTool, 'onMigrate' )

    def tool_list( self ):
        print("List of tools supported by CID:")
        for k in sorted(self._tool_impl.keys()):
            t = self._tool_impl[k]
            doc = t.__doc__.strip() or '!! Missing documentation.'
            doc = doc.split("\n")[0]
            print("  * " + k + ': ' + doc)
        print('End.')

    def tool_describe( self, tool ):
        t = self._tool_impl[tool]
        
        print('* Tool: ' + tool)
        
        env_vars = t.envNames()
        if env_vars:
            print('* Environment variables: ' + ', '.join(env_vars))

        deps = t.getDeps()
        if deps:
            print('* Dependencies: ' + ', '.join(deps))

        rdeps = t.getPostDeps()
        if rdeps:
            print('* Reverse dependencies: ' + ', '.join(rdeps))
            
        order = t.getOrder()
        if order:
            print('* Order shift: {0}'.format(order))

        print()
        doc = t.__doc__.strip() or '!! Missing documentation.'
        print(doc)
        print()

    def _initConfig( self ):
        self._global_config = gc = self._loadJSON( '/etc/futoin/futoin.json', {'env':{}} )
        self._user_config = uc = self._loadJSON( os.path.join( os.environ.get('HOME','/'), 'futoin.json' ), {'env':{}} )
        self._project_config = pc = self._loadJSON( 'futoin.json', {} )

        deploy_dir = self._overrides.get( 'deployDir', None )
        dc = {'env':{}}
        if deploy_dir :
            dc = self._loadJSON( os.path.join( deploy_dir, 'futoin.json' ), dc )
        self._deploy_config = dc
        
        merged = OrderedDict( pc )
        
        if 'env' in pc:
            raise RuntimeError( '.env node is set in project config' )

        if 'env' not in dc or len( dc ) != 1:
            raise RuntimeError( 'Deploy config must have only .env node' )
        
        if 'env' not in uc or len( uc ) != 1:
            raise RuntimeError( 'User config must have only .env node' )
        
        if 'env' not in gc or len( gc ) != 1:
            raise RuntimeError( 'Glboal config must have only .env node' )
        
        env = OrderedDict( dc['env'] )
        
        for ( k, v ) in uc.items():
            env.setdefault( k, v )
        for ( k, v ) in gc.items():
            env.setdefault( k, v )
        
        self._initEnv( env )
        merged['env'] = env
        merged.update( self._overrides )
        self._config = merged
        
        self._initTools()
    
    def _loadJSON( self, file_name, defvalue ):
        try :
            with open(file_name, 'r') as content_file:
                content = content_file.read()
                object_pairs_hook = lambda pairs: OrderedDict( pairs )
                return json.loads( content, object_pairs_hook=object_pairs_hook )
        except IOError:
            return defvalue
        
    def _initEnv( self, env ) :
        env.setdefault( 'type', 'dev' )
        env.setdefault( 'startup', 'cron' )
        env.setdefault( 'webServer', 'nginx' )
        env.setdefault( 'vars', {} )
        env.setdefault( 'mainConfig', {} )
        env.setdefault( 'plugins', {} )

        env.setdefault( 'externalSetup', {} )
        externalSetup = env['externalSetup']
        externalSetup.setdefault( 'runCmd', None )
        externalSetup.setdefault( 'webServer', False )
        externalSetup.setdefault( 'startup', False )
        externalSetup.setdefault( 'installTools', False )

        env.setdefault( 'binDir', os.path.join(os.environ['HOME'], 'bin') )
        self._addBinPath( env['binDir'] )

    def _initTools( self ):
        config = self._config
        env = config['env']

        #---
        tool_impl = self._tool_impl
        
        if tool_impl is None:
            tool_impl = {}
            default_tool_dir = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'tool'
            )
            
            tool_files = os.listdir(default_tool_dir)
            tool_files = fnmatch.filter(tool_files, '*tool.py')

            for f in tool_files:
                k = f.replace('tool.py', '')
                pkg = 'futoin.cid.tool.' + k + 'tool'
                m = importlib.import_module( pkg )
                tool_impl[ k ] = getattr( m, k + 'Tool' )( k )

            self._tool_impl = tool_impl
            
        plugins = env['plugins'].copy()
        plugins.update( config.get( 'plugins', {} ) )
        
        for ( k, v ) in plugins.items() :
            if k not in tool_impl:
                m = importlib.import_module( v )
                tool_impl[ k ] = getattr( m, k + 'Tool' )( k )

        #---
        curr_tool = config.get('tool', None)
        
        if curr_tool:
            tools = [ curr_tool ]
        else :
            tools = config.get('tools', [])
            
            if not tools:
                for ( n, t ) in tool_impl.items():
                    if t.autoDetect( config ) :
                        tools.append( n )

            # Make sure deps & env are processed for cli-supplied tools
            #--
            for (item, base) in {'rms' : RmsTool, 'vcs' : VcsTool}.items():
                tool = config.get(item, None)

                if tool:
                    tools.append( tool )

                    if not isinstance(tool_impl[ tool ], base):
                        raise RuntimeError('Tool {0} does not suite {1} type'.format(tool, item))

        # add all deps
        #--
        dep_generations = [ set(tools) ]
        tools = set( tools )
        postdeps = set()
        dep_length = 0
        last_index = 0
        while len( dep_generations ) != dep_length :
            dep_length = len( dep_generations )
            curr_index = last_index
            last_index = len(dep_generations)

            for g in dep_generations[curr_index:]:
                for tn in g:
                    t = tool_impl[tn]
                    moredeps = set(t.getDeps())
                    if moredeps:
                        dep_generations.append( moredeps )
                        tools.update( moredeps )
                    postdeps.update( set(t.getPostDeps()) - tools )

            if len( dep_generations ) == dep_length and postdeps:
                dep_generations.append( postdeps )
                tools.update( postdeps )
                postdeps = set()
    
        #---
        dep_generations.reverse()
        tools = []
        for d in dep_generations :
            tools.extend( d - set(tools) )
        config['tools'] = tools

        #--
        if config['toolTest']:
            for tool in tools :
                t = tool_impl[tool]
                t.importEnv( env )
                t.initEnv( env )
        else :
            for tool in tools :
                t = tool_impl[tool]
                t.requireInstalled( env )
                if tool != curr_tool:
                    t.loadConfig( config )

        # Solves generic issues of ordering independent tools in
        # later execution with predictable results:
        # 1. sort by integer order
        # 2. sort by tool name
        tools.sort( key=lambda v: ( tool_impl[v].getOrder(), v ) )




from ..buildtool import BuildTool
from ..runenvtool import RunEnvTool
from .piptoolmixin import PipToolMixIn

class dockercomposeTool( PipToolMixIn, BuildTool, RunEnvTool ):
    """Compose is a tool for defining and running multi-container Docker applications.
    
Home: https://docs.docker.com/compose/

Experimental support.
"""
    def autoDetect( self, config ) :
        return self._autoDetectByCfg(
                config,
                [ 'docker-compose.yml', 'docker-compose.yaml' ]
        )
    
    def getDeps( self ) :
        return [ 'pip', 'docker' ]
    
    def getOrder( self ):
        return 20
        
    def _pipName( self ):
        return 'docker-compose'
    
    def _installTool( self, env ):
        self._requirePythonDev( env )
        super(dockercomposeTool, self)._installTool( env )

    def onBuild( self, config ):
        self._callExternal( [ config['env']['dockercomposeBin'], 'build' ] )
        
    def initEnv( self, env ) :
        super(dockercomposeTool, self).initEnv( env, 'docker-compose' )
